/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  hashGenerator.cpp
///  \brief Implements some primitive methods to get well-defined hash values.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   2010-09-29
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#include "hashGenerator.h"

std::vector<uchar> HashGenerator::hash_table;

std::vector<uchar>::const_iterator HashGenerator::begin(void)
{
	return hash_table.begin();
}

void HashGenerator::resize(unsigned int size)
{
	if(hash_table.size()>=size)
	{
		return;
	}

	hash_table.resize(size, 0);
	uchar  x( 3767350049ULL);
	uchar  y(1394464459ULL);

	for(unsigned int i=0; i<size; ++i)
	{
		hash_table[i] = (x^y)+x;
		x *= (x-y);
		y ^= x;
	}
}

uchar HashGenerator::getHash(const std::vector<unsigned int>& vals,
		const std::vector<unsigned int>::const_iterator& blockIterBeg, const std::vector<unsigned int>::const_iterator& blockIterEnd )
{
	uchar res(0);
	std::vector<unsigned int>::const_iterator valsIter(vals.begin());
	std::vector<uchar>::const_iterator hashIter (hash_table.begin()+2);
	std::vector<unsigned int>::const_iterator blockIter(blockIterBeg);

	for(; blockIter<blockIterEnd; ++blockIter, ++valsIter)
	{
		res ^= *(hashIter + *valsIter);
		hashIter+= (*blockIter)+1;
	}
	return res;
}

uchar HashGenerator::getHash(const std::vector<std::vector<unsigned int> >& vals)
{
    uchar res(0);
    std::vector<std::vector<unsigned int> >::const_iterator valsIter(vals.begin());
    std::vector<std::vector<unsigned int> >::const_iterator valsIterEnd(vals.end());
    std::vector<uchar>::const_iterator hashIter (hash_table.begin());

    for(; valsIter<valsIterEnd; ++valsIter)
    {
        std::vector<unsigned int>::const_iterator valsEntIter(valsIter->begin());
        std::vector<unsigned int>::const_iterator valsEntIterEnd(valsIter->end());

        for( ; valsEntIter<valsEntIterEnd; valsEntIter++, hashIter++)
        {
            res ^= *hashIter * (*valsEntIter);
        }
    }
    return res;
}
