/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  magmaCASModule.cpp
///  \brief Provides an interface to MAGMA.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   2010-09-29
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#include "magmaCASModule.h"
#include "actingElement.hpp"
#include "groups.h"
//#include "z4.hpp"
//#include "f2_f2.hpp"
#include <sys/types.h>
#include <sys/wait.h> 

bool MagmaCASModule::verifyMinDist(unsigned int d, unsigned int dual_d)
{
    m_to<<"C := LinearCode(Gamma);"<<std::endl;
    m_to<<"MinimumDistance(C);"<<std::endl;
    unsigned int res;
    m_from>>res;
    if(res<d)
    {
        return false;
    }
    m_to<<"MinimumDistance(Dual(C));"<<std::endl;
    m_from>>res;
    if(res<dual_d)
    {
        return false;
    }
    return false;
}

template<>
bool MagmaCASModule::verifyMinDist(const Matrix<Field>& matrix, unsigned int d, unsigned int dual_d)
{
    char sep = ',';
    m_to<<"Gamma := Matrix("<<m_ringname<<", "<<matrix.getRows()<<","<<matrix.getCols()<<", [";
    matrix.print(m_to, &sep, &m_toElt);
    m_to<<"]);"<<std::endl;

    return verifyMinDist(d, dual_d);
}


template<>
bool MagmaCASModule::verifyMinDist(const Matrix<BinaryField>& matrix, unsigned int d, unsigned int dual_d)
{
    m_to<<"Gamma := Matrix(GF(2), "<<matrix.getRows()<<","<<matrix.getCols()<<", [";
    matrix.print(m_to, ",", NULL);
    m_to<<"]);"<<std::endl;

    return verifyMinDist(d, dual_d);
}

//
//template<>
//unsigned int MagmaCASModule::dualMinDist<1u>(const unsigned int* const type, unsigned int n, unsigned int q,
//        const Distribution<SymWeight<1u> >& primal_wd)
//{
//    m_to<<"MacWilliamsTransform("<<n<<","<<type[0]<<","<<q<<","<<"Sort([";
//
//    std::map<SymWeight<1u>, unsigned int>::const_iterator wdIter(primal_wd.begin());
//    std::map<SymWeight<1u>, unsigned int>::const_iterator wdIterEnd(primal_wd.end());
//
//    if(wdIter!=wdIterEnd)
//    {
//        m_to<<"<"<<wdIter->first.qty<<", "<<wdIter->second<<">";
//        ++wdIter;
//        for( ; wdIter!=wdIterEnd; ++wdIter)
//        {
//           m_to<<", <"<<wdIter->first.qty<<", "<<wdIter->second<<">";
//        }
//    }
//
//    m_to<<"]))[2][1];"<<std::endl;
//
//    unsigned int d;
//    m_from>>d;
//    return d;
//}
//
//template<>
//unsigned int MagmaCASModule::dualMinDist<2u>(const unsigned int* const type, unsigned int n, unsigned int q,
//        const Distribution<SymWeight<2u> >& primal_wd)
//{
//    unsigned int* gray_type = new unsigned int[1];
//    gray_type[0] = type[0]*2 + type[1];
//    unsigned int d( dualMinDist<1u>( gray_type, 2*n, 2, primal_wd.GrayImageDistribution()));
//    delete[] gray_type;
//    return d;
//}


MagmaCASModule::MagmaCASModule(char* call[], const std::string& packagesFolder) : AbstractCASModule(call), m_ringname("RRR"), m_toElt("toField"),
        m_toInt("fromField")
{
    m_to << "load \""<< packagesFolder <<"labra.magma\";;"<<std::endl;

    m_to<<"if assigned getFather then"<<std::endl;
    m_to << "print \"ready\";"<<std::endl;
    m_to << "else;"<<std::endl;
    m_to << "print \"load failed\";"<<std::endl;
    m_to << "end if; "<<std::endl;

    char x [256];
    while(m_from.getline(x,256) )
    {
        if(!strcmp(x, "ready"))
            break;
        if(!strcmp(x, "load failed"))
        {
            std::cerr<<"Problems loading magma"<<std::endl<<"Exiting."<<std::endl;
            exit(-1);
        }
    }
}
MagmaCASModule::~MagmaCASModule(void)
{
    m_to<<"exit;"<<std::endl;
}

void MagmaCASModule::getFather(const std::string& group, std::vector<int>& father, My_Long_Int& group_order )
{
    m_to<<"getFather("<<group<<", "<<father.size()<<");"<<std::endl;
    std::vector<int>::const_iterator endIter = father.end();
    for(std::vector<int>::iterator iter=father.begin(); iter != endIter; ++iter)
    {
        m_from>> *iter;
    }
    m_from>>group_order;
}

void MagmaCASModule::initTrivialSn(const std::string& group, int n)
{
    m_to<<group<<" := sub<Sym( "<<n<<")| >;"<<std::endl;
}

void MagmaCASModule::extendGroup(const std::string& group, const Permutation& newgen, int n)
{
    m_to<<"S := Sym("<<n<<");"<<std::endl;
    m_to<<"new := sub<S | S!";
    newgen.print(m_to,1, Permutation::cycle_notation);
    m_to<<">;"<<std::endl;
    m_to<<group<<" := sub<S |  ["<<group<<", new]>;"<<std::endl;
}

void MagmaCASModule::buildGroup(const std::string& group, const std::vector<Permutation>& newgens, int n)
{
    m_to<<"S := Sym("<<n<<");"<<std::endl;
    m_to<<"gen := [S |   ];"<<std::endl;
    for(unsigned int i=0; i<newgens.size(); ++i)
    {
        m_to<<"Append(~gen, S!";
        newgens[i].print(m_to,1, Permutation::cycle_notation);
        m_to<<");"<<std::endl;
    }
    m_to<<group<<" := sub<S |  gen>;"<<std::endl;
}


My_Long_Int MagmaCASModule::getOrder( const std::string& group )
{
    My_Long_Int order(0);
    m_to<<"Order("<<group<<");"<<std::endl;
    m_from>>order;

    return order;
}

unsigned int MagmaCASModule::BKLCUpperBound( unsigned int n, unsigned int k, unsigned int q )
{
    if( q > 9)
    {
        return n;
    }


    m_to<<"BKLCUpperBound(GF("<<q<<"), "<<n<<","<<k<<");"<<std::endl;
    unsigned int d(0);
    m_from>>d;
    return d;
}

unsigned int MagmaCASModule::BKLCLowerBound( unsigned int n, unsigned int k, unsigned int q )
{
    if( q > 9)
    {
        return 1;
    }

    m_to<<"BKLCLowerBound(GF("<<q<<"), "<<n<<","<<k<<");"<<std::endl;
    unsigned int d(0);
    m_from>>d;
    return d;
}


void MagmaCASModule::fillGLsubgroup( unsigned int k, const std::vector<ActingElement<Field, PLUSMINUSMONOMIAL> >& autom_gens,
        const std::string& groupname )
{
    std::vector<ActingElement<Field, PLUSMINUSMONOMIAL> >::const_iterator autom_gensIter(autom_gens.begin());
    std::vector<ActingElement<Field, PLUSMINUSMONOMIAL> >::const_iterator autom_gensIterEnd(autom_gens.end());

    m_to<<groupname<<":= sub<GeneralLinearGroup("<<k<<", "<<m_ringname<<") | ["<<std::endl;
    do
    {
        m_to<<"Transpose(Matrix("<<m_ringname<<"," << k<<", "<<k<< ", [";
        autom_gensIter->m_A.print(m_to, ",", &m_toElt);
        m_to<<"]))";
        if(++autom_gensIter != autom_gensIterEnd)
        {
            m_to<<", "<<std::endl;
        }
        else
        {
            break;
        }
    }while(true);
    m_to<<"]>;"<<std::endl;
}

void MagmaCASModule::calcStabilizer(const std::string& groupname, const std::string& stabname, const Vec<Field>& v)
{
    m_to<<stabname<<" := Stabilizer("<<groupname<<", Vector( ["<<m_ringname<<"| "<<std::endl;
    v.print(m_to, m_toElt);
    m_to<<"]));"<<std::endl;
}

template<>
void MagmaCASModule::calcOrbitReps(unsigned int k, const std::string& groupname,  std::vector<Vec<Field > >& orbitReps )
{
    m_to<<"orb := Orbits("<<groupname<<");"<<std::endl;
    m_to<<"print #orb;"<<std::endl;
    unsigned int size;
    m_from>>size;
    m_to<<"for o in orb do"<<std::endl;
    m_to<<"s :=  ElementToSequence(Minimum(o));"<<std::endl;
    m_to<<"for x in s do"<<std::endl;
    m_to<<"print "<<m_toInt<<"(x);"<<std::endl;
    m_to<<"end for;"<<std::endl;
    m_to<<"end for;"<<std::endl;

    orbitReps = std::vector<Vec<Field> >(size, Vec<Field>(k));
    std::vector<Vec<Field> >::iterator orbitRepIter(orbitReps.begin());
    std::vector<Vec<Field> >::iterator orbitRepIterEnd(orbitReps.end());
    unsigned int dummy;

    for( ; orbitRepIter<orbitRepIterEnd; ++orbitRepIter)
    {
        Vec<Field>::iterator vIter(orbitRepIter->begin());
        Vec<Field>::iterator vIterEnd(orbitRepIter->end());
        for(; vIter<vIterEnd; ++vIter)
        {
            m_from>>dummy;
            *vIter = Element<Field>(dummy);
        }
    }
}

template<>
void MagmaCASModule::initRingStructure<Z4>(void)
{
    std::cout<<"not implemented, yet";
    exit(-1);
}

template<>
void MagmaCASModule::initRingStructure<F2_F2>(void)
{
    std::cout<<"not implemented, yet";
    exit(-1);
}

template<>
void MagmaCASModule::initRingStructure<BinaryField>(void)
{
    m_to<<m_ringname<<" := GF(2);"<<std::endl;
    m_to<<m_toElt<<":= map< {0 ,1} -> "<<m_ringname<<" | x :-> x>;"<<std::endl;
    m_to<<m_toInt <<":= map<"<<m_ringname<<" -> {0,1} | x :->Integers()!x>;"<<std::endl;
}

template<>
void MagmaCASModule::initRingStructure<Field>(void)
{
    unsigned int p(Field::get_primefield_size());
    unsigned int r(Field::get_r());
    if(r==1)
    {
        m_to<<m_ringname<<" := GF("<<p<<");"<<std::endl;
        m_to<<m_toElt<<":= map< {0 .. "<< p-1<<"} -> "<<m_ringname<<" | x :-> "<<m_ringname<<"!x>;"<<std::endl;
        m_to<<m_toInt <<":= map<"<<m_ringname<<" -> {0 .. "<< p-1<<"} | x :->Integers()!x>;"<<std::endl;
    }
    else
    {
        Element<Field> xi( Field::get_primitiveElement());
        Element<Field> res(xi);
        for(unsigned int i=2; i<=r; ++i)
        {
            res *= xi;
        }

        m_to<<m_ringname<<"<xi> := ExtensionField<GF("<<p<<"), x | x^"<<r;
        std::vector<unsigned int> basisRep( res.getBasisRepresentation() );
        for(unsigned int i=0; i<r; ++i)
        {
            m_to<<"-"<<basisRep[i]<<"*x^"<<i<<std::endl;
        }
        m_to<<">;"<<std::endl;

        m_to<<"mapping := [<0, "<<m_ringname<<"!0>";
        for(unsigned int i=1; i<Field::my_size(); ++i)
        {
            std::vector<unsigned int> basisRep( Element<Field>(i).getBasisRepresentation() );

            m_to<<", <"<<i<<","<<basisRep[0];
            for(unsigned int j=1; j<r; ++j)
            {
                m_to<<"+ "<<basisRep[j]<<"*xi^"<<j;
            }
            m_to<<">"<<std::endl;
        }
        m_to<<"];"<<std::endl;


        m_to<<m_toElt<<" := map< {0 .. "<<Field::my_size()-1<<"} -> "<<m_ringname<<" | mapping>;"<<std::endl;
        m_to<<m_toInt<<" := map< "<<m_ringname<<" -> {0 .. "<< Field::my_size()-1<<"} | [ <m[2], m[1]> :  m in mapping] >;"<<std::endl;
    }
}



//unsigned long MagmaCASModule::getMinWords(const Matrix<Z4>& matrix, vector<Vec<Z4> >& minWords)
//{
//	minWords.clear();
//	m_to<<"Gamma := Matrix(Integers(4), "<<matrix.getRows()<<","<<matrix.getCols()<<", [";
//	matrix.print(m_to, ",");
//	m_to<<"]);";
//	m_to<<"calcMinWordSet( Gamma );"<<endl;
//	unsigned long qty;
//	m_from>>qty;
//
//	int dummy;
//	Vec<Z4> cw(matrix.getCols(),0);
//	for(unsigned long i=0; i<qty; ++i)
//	{
//		for(unsigned j=0; j<matrix.getCols(); ++j)
//		{
//			m_from>>dummy;
//			cw[j] = dummy;
//		}
//		minWords.push_back(cw);
//	}
//	return qty;
//}

